@@ -2,10 +2,10 @@
 #
 # command.tcl -- make gdb commands happen under tgdb's control
 #
-# (c) 1994 HighTec EDV-Systeme GmbH
-#          Neue Bahnhofstr. 71
-#          66386 St.Ingbert, Germany
-#          Email: tgdb@HighTec.saarlink.de
+# (c) 1994-95 HighTec EDV-Systeme GmbH
+#             Feldmannstr. 98
+#             66119 Saarbruecken, Germany
+#             Email: tgdb@HighTec.saarlink.de
 #
 # *** ALL RIGHTS RESERVED ***
 #
@@ -17,23 +17,32 @@
 # use the file command for gdb 3.x, too :-)
 ################################################################################
 proc t_file { {file ""} } {
-  global Tgdb_interactive QueryResult debugger SourcePath
+  global Tgdb_interactive QueryResult debugger gdb16x SourcePath Tgdb_option
 
   t_exec_file $file
   if { $QueryResult != "n" } {
     t_symbol_file $file
+  } else {
+    return
   }
-  if { ($file != "") && ($debugger == "gdb166") } {
+  if { ($file != "") && ($debugger == "gdb166") && !$gdb16x } {
     set path [file dirname $file]
     canonicalize path
+    set psep ":"
+    if { [string first ";" $SourcePath] != -1 } {
+      set psep ";"
+    }
     if {   ($path != "") && ($path != ".")
-        && ([lmatch [split $SourcePath ":"] $path] == "")} {
+        && ([lmatch [split $SourcePath $psep] $path] == "")} {
       set tmp_interactive $Tgdb_interactive
       set Tgdb_interactive 0
       t_directory "$path"
       set Tgdb_interactive $tmp_interactive
     }
   }
+  if { $Tgdb_option(ProgPath) && ($file != "") } {
+    t_cd [file dirname $file]
+  }
   do_dialog "set \$pc=main" silent
   t_list "main,"
 }
@@ -240,7 +249,9 @@ proc t_exec_file { {file ""} } {
   } else {
     append_to_gdbwin "exec-file $file\n$result$prompt"
   }
-  update_windows new
+  if { ![cequal $result "Program not killed.\n"] } {
+    update_windows new
+  }
 }
 ################################################################################
 # realize gdb's symbol-file command
@@ -254,11 +265,15 @@ proc t_symbol_file { {file ""} } {
   } else {
     append_to_gdbwin "symbol-file $file\n$result$prompt"
   }
-  if { !$Tgdb_option(UpdateBpts) } {
-    get_all_disps
-    update_bpts
+  if { [regexp {^Reading symbols from.*} $result] } {
+    init_srcwin
+    t_list "main,"
+    if { !$Tgdb_option(UpdateBpts) } {
+      get_all_disps
+      update_bpts
+    }
+    update_windows new
   }
-  update_windows new
 }
 ################################################################################
 # realize gdb's core-file command
@@ -394,16 +409,16 @@ proc t_source { {script ""} } {
 # to simulate the -fullname effects of 4.x
 ################################################################################
 proc t_info_line { {which ""} {how verbose} {showpc no} {loadfile yes} } {
-  global debugger
+  global debugger gdb16x
 
   set result [do_dialog "info line $which" $how $showpc $loadfile]
-  if { $debugger != "gdb166" } {return $result}
+  if { ($debugger != "gdb166") || $gdb16x } {return $result}
   # simulate gdb 4.x behaviour...
   if { ![regexp {^Line ([0-9]+) of "(.*)" .* at (pc|address) (0x[0-9a-f]+)} \
 	  $result {} line file fill addr] } {
     return $result
   }
-  if { [cequal [set file [file2path $file $line]] $file] } {
+  if { [cequal $file [set file [file2path $file $line]]] } {
     return $result
   }
   if { $loadfile != "yes" } {
@@ -508,11 +523,11 @@ proc t_list { {what ""} } {
 # source path; we're really interested in what's goin' on there...
 ################################################################################
 proc t_directory { {dir ""} } {
-  global Tgdb_interactive SourcePath PathCache debugger
+  global Tgdb_interactive SourcePath PathCache debugger gdb16x
 
   canonicalize dir
   if { $dir == "" } {
-    if { $debugger == "gdb166" } {
+    if { ($debugger == "gdb166") && !$gdb16x } {
       set result [do_dialog "info directories" silent]
       regexp {Source directories searched: (.*)} $result {} old_path
     } else {
@@ -526,12 +541,53 @@ proc t_directory { {dir ""} } {
     set result [do_dialog "directory $dir" silent]
   }
   regexp {Source directories searched: (.*)} $result {} SourcePath
+  if { !$Tgdb_interactive } {
+    show_status $result steady
+  }
   if { ($dir == "") && ([string compare $SourcePath $old_path] != 0) } {
     unset PathCache
   }
   set SourcePath [string trim $SourcePath]
 }
 ################################################################################
+# change current working directory
+################################################################################
+proc t_cd { {dir ""} } {
+  global Tgdb_interactive SourcePath PathCache debugger gdb16x
+
+  if { $dir != "" } {
+    catch {cd $dir}
+  }
+  if { $Tgdb_interactive } {
+    do_dialog "cd $dir" verbose
+  } else {
+    show_status [do_dialog "cd $dir" silent] steady
+  }
+}
+################################################################################
+# examine memory
+# !!! must only be called interactively (i.e. Tgdb_interactice = 1) !!!
+################################################################################
+proc t_x { {arg ""} } {
+  global Last_x_format
+
+  if { ![info exists Last_x_format] } {
+    set Last_x_format ""
+  }
+  if { [cequal $arg "?noarg?"] } {
+    do_dialog "x $Last_x_format \$last_x_address" verbose
+  } else {
+    if { [string match "/*" $arg] } {
+      do_dialog "x $arg" verbose
+      set Last_x_format [lindex $arg 0]
+    } else {
+      do_dialog "x $Last_x_format $arg" verbose
+    }
+  }
+  do_dialog "x/1" silent
+  do_dialog "set \$last_x_address = \$_" silent
+}
+################################################################################
 # just a helper for t_directory: make dots and stuff disappear from the
 # path name so that it finally becomes a proper one...
 ################################################################################
@@ -1039,7 +1095,9 @@ proc update_stack_window { {how new} } {
     .stack.f0.lb select from $fno
     # make the selected item visible within the listbox
     # (similar to "yview -pickplace pos" for text widgets)
-    set h [expr [winfo height .stack.f0.lb] / $WinSize(.stack,y)]
+    set h [expr \
+      ([winfo height .stack.f0.lb] - [lindex [.stack.f0.lb conf -bd] 4]) \
+      / $WinSize(.stack,y)]
     set i0 [.stack.f0.lb nearest 0]
     if { ($fno < $i0) || ($fno >= [expr $i0 + $h - 1]) } {
       .stack.f0.lb yview [expr $fno - 2]
@@ -1318,7 +1376,7 @@ proc update_asm_window { expr {rebuild no} } {
       }
     }
   }
-  set h [expr [winfo height $w] / $WinSize(.asm,y)]
+  set h [expr ([winfo height $w] - [lindex [$w conf -bd] 4]) / $WinSize(.asm,y)]
 
   if { $rebuild == "no" } {
     # check if $start_pc is visible within text widget...
@@ -1527,6 +1585,22 @@ proc update_disp_window { {result ""} } {
   }
   .disp.f0.lb yview $FirstPos(.disp)
 }
+################################################################################
+# realize pxmon commands
+################################################################################
+proc t_pxmon { {cmd ""} } {
+  global WinPos
+
+  if { ($cmd == "off") || ($cmd == "x") } {
+    if { [winfo exists .pxmon] } {
+      set WinPos(.pxmon) [wm geometry .pxmon]
+      destroy .pxmon
+    }
+    show_status "Continuing..."
+    append_to_gdbwin "Continuing application ...\n"
+  }
+  do_dialog "pxmon $cmd" verbose yes
+}
 
 ################################################################################
 #
@@ -1619,10 +1693,10 @@ proc t_break { {bpexpr ""} } {
 # it for deletion (which I emulate for gdb 3.x for now).
 ################################################################################
 proc t_tbreak { {bpexpr ""} } {
-  global debugger Tgdb_interactive prompt
+  global debugger Tgdb_interactive prompt gdb16x
 
   set retval 0
-  if { $debugger == "gdb166" } {
+  if { ($debugger == "gdb166") && !$gdb16x } {
     set result [do_dialog "break $bpexpr" silent]
     if { [regexp {Breakpoint [0-9]* at (0x[0-9a-f]*)} $result] } {
       set retval 1
@@ -1778,7 +1852,7 @@ proc read_source_file {} {
 #
 ################################################################################
 proc write_init_file { file } {
-  global debugger gdb_class gdb_cmd
+  global debugger gdb_class gdb_cmd gdb16x
 
   if { $file == "" } {
     show_status "Command cancelled."
@@ -1800,7 +1874,7 @@ proc write_init_file { file } {
   if { [regexp {^Working directory (.*)\.} $result {} result] } {
     puts $desc "cd $result"
   }
-  if { $debugger == "gdb166" } {
+  if { ($debugger == "gdb166") && !$gdb16x } {
     set result [do_dialog "info directories" silent]
   } else {
     set result [do_dialog "show directories" silent]
@@ -1810,7 +1884,7 @@ proc write_init_file { file } {
   }
 
   set result [do_dialog "info files" silent]
-  if { $debugger == "gdb166" } {
+  if { ($debugger == "gdb166") && !$gdb16x } {
     if { [regexp {Executable file "([^"]*)"} $result {} exec] } {
       puts $desc "exec-file $exec"
     }
@@ -1826,7 +1900,7 @@ proc write_init_file { file } {
     puts $desc "symbol-file $symbol"
   }
   puts $desc ""
-  if { $debugger == "gdb166" } {
+  if { ($debugger == "gdb166") && !$gdb16x } {
     set bpinfo [split [do_dialog "info breakpoints" silent] "\n"]
     if { [string match "Breakpoints:" [lvarpop bpinfo]] } {
       lvarpop bpinfo
@@ -1943,7 +2017,7 @@ proc write_init_file { file } {
   find_gdb_cmds "user" "user"
   if { [info exists gdb_class(user)] } {
     foreach ucmd $gdb_class(user) {
-      if { $debugger == "gdb166" } {
+      if { ($debugger == "gdb166") && !$gdb16x } {
 	set result [do_dialog "list %-$ucmd" silent]
         regsub -all {[ ]*[0-9]+: } $result "" result
       } else {
@@ -2008,10 +2082,14 @@ proc expr2line { expr } {
       set result [t_info_line "$expr" silent no no]
     }
   }
-  if { ![regexp {([^:]*):([0-9]+):([0-9]+):(beg|middle):(0x[0-9a-f]+)} \
-          $result {} file line off pc] } {
+  if { ![regexp \
+       {((.:)?[^:]*)[:;]([0-9]+)[:;]([0-9]+)[:;](beg|middle):(0x[0-9a-f]+)}\
+          $result {} file {} line off pc] } {
     return ""
   }
+  if { ![file readable $file] } {
+    set file [file2nfspath $file]
+  }
   return "$file $line"
 }
 ################################################################################
@@ -2033,8 +2111,9 @@ proc expr2info { expr } {
       set result [t_info_line "$expr" silent no no]
     }
   }
-  if { ![regexp {([^:]*):([0-9]+):([0-9]+):(beg|middle):(0x[0-9a-f]+)} \
-          $result {} file line off pc] } {
+  if { ![regexp \
+       {((.:)?[^:]*)[:;]([0-9]+)[:;]([0-9]+)[:;](beg|middle):(0x[0-9a-f]+)}\
+          $result {} file {} line off pc] } {
     return ""
   }
   return "$file $line $pc"
@@ -2166,13 +2245,17 @@ proc get_all_disps {} {
 # minimize accesses to gdb
 ################################################################################
 proc file2path { file line } {
-  global PathCache SourcePath debugger
+  global PathCache SourcePath debugger gdb16x
 
   if { [info exists PathCache($file)] } {
     return $PathCache($file)
   }
-  if { $debugger == "gdb166" } {
-    foreach dir [split $SourcePath ":"] {
+  set psep ":"
+  if { [string first ";" $SourcePath] != -1 } {
+    set psep ";"
+  }
+  if { ($debugger == "gdb166") && !$gdb16x } {
+    foreach dir [split $SourcePath $psep] {
       if { [file exists "$dir/$file"] } {
 	return [set PathCache($file) "$dir/$file"]
       }
@@ -2186,6 +2269,26 @@ proc file2path { file line } {
   }
 }
 ################################################################################
+# Given a file (which might be an absolute path), return file's full path; we
+# use a cache to minimize accesses to gdb
+# This function is needed if directories are mounted via NFS.
+################################################################################
+proc file2nfspath { file } {
+  global PathCache SourcePath
+
+  set psep ":"
+  if { [string first ";" $SourcePath] != -1 } {
+    set psep ";"
+  }
+  set filename [file tail $file]
+  foreach dir [split $SourcePath $psep] {
+    if { [file exists "$dir/$filename"] } {
+      return [set PathCache($file) "$dir/$filename"]
+    }
+  }
+  return $file
+}
+################################################################################
 # Given a breakpoint number, return corresponding address in the form
 # "path lineno" (or "" if that fails).
 ################################################################################
@@ -2243,14 +2346,14 @@ proc expr2bpno { expr } {
 ################################################################################
 proc find_bpt_cmds { num } {
   upvar $num bpnum
-  global debugger
+  global debugger gdb16x
 
-  if {   (($debugger == "gdb166") && ($bpnum == "\$"))
-      || (($debugger != "gdb166") && ($bpnum == "\$bpnum")) } {
+  if {   (($debugger == "gdb166") && !$gdb16x && ($bpnum == "\$"))
+      || ((($debugger != "gdb166") || $gdb16x) && ($bpnum == "\$bpnum")) } {
     set bpnum ""
   }
   set bpinfo [split [do_dialog "info breakpoints $bpnum" silent] "\n"]
-  if { $debugger == "gdb166" } {
+  if { $debugger == "gdb166" && !$gdb16x } {
     set line [lindex $bpinfo 0]
     if { [string compare "Breakpoints:" $line] && ![regexp {^#[0-9]+} $line] } {
       return "false"
@@ -2468,6 +2571,7 @@ proc init_tgdb_cmds {} {
   set Tgdb_cmd([list disable breakpoints]) t_disable
   set Tgdb_cmd(break) t_break
   set Tgdb_cmd(tbreak) t_tbreak
+  set Tgdb_cmd(cd) t_cd
   set Tgdb_cmd(clear) t_clear
   set Tgdb_cmd(directory) t_directory
   set Tgdb_cmd(shell) t_shell
@@ -2507,6 +2611,8 @@ proc init_tgdb_cmds {} {
   set Tgdb_cmd([list show history size]) t_show_history_size
   set Tgdb_cmd([list show commands]) t_show_commands
   set Tgdb_cmd([list info commands]) t_show_commands
+  set Tgdb_cmd(pxmon) t_pxmon
+  set Tgdb_cmd(x) t_x
 
   ##############################################################################
   # for debugging purposes only: delete a given procedure; it will be reloaded
