@@ -3,16 +3,16 @@
 # tgdb - A graphical frontend to gdb, the GNU debugger.
 #	 Based on Tcl/Tk, TclX, TkSteal, Expect and BLT.
 #
-# (c) 1994 HighTec EDV-Systeme GmbH
-#          Neue Bahnhofstr. 71
-#          66386 St.Ingbert, Germany
-#          Email: tgdb@HighTec.saarlink.de
+# (c) 1994-95 HighTec EDV-Systeme GmbH
+#             Feldmannstr. 98
+#             66119 Saarbruecken, Germany
+#             Email: tgdb@HighTec.saarlink.de
 #
 # *** ALL RIGHTS RESERVED ***
 #
 ################################################################################
 
-set tgdb_version 1.1
+set tgdb_version 1.4
 wm withdraw .
 focus none
 catch {
@@ -28,6 +28,7 @@ catch {
   
 catch {selection clear .}
 set appname [winfo name .]
+set Tgdb_cwd [pwd]
 set std_out [dup stdout]
 close stdout
 ################################################################################
@@ -38,6 +39,13 @@ close stdout
 if { [catch {set debugger $env(TGDB_DEBUGGER)}] } {
   set debugger gdb
 }
+# see if this is gonna be a remote session
+if { [cequal [lindex $debugger 0] rsh] } {
+  set dbgrsh [lrange $debugger 0 [expr [llength $debugger] - 2]]
+  set debugger [lindex $debugger [expr [llength $debugger] - 1]]
+} else {
+  set dbgrsh ""
+}
 if { [catch {set prompt $env(TGDB_PROMPT)}] } {
   set prompt "($debugger) "
 }
@@ -57,6 +65,11 @@ if { [info exists env(TGDB_SMALLICONS)] && ($env(TGDB_SMALLICONS) == "1") } {
 if { [catch {set tgdb_appdef_path $env(TGDB_APPDEF_PATH)}] } {
   set tgdb_appdef_path "."
 }
+if { ($debugger == "gdb166") && ($prompt == "(gdb16x) ") } {
+  set gdb16x 1
+} else {
+  set gdb16x 0
+}
 ################################################################################
 # examine the argument(s) which we should pass to gdb; do this now,
 # so that we can quickly exit if things go wrong
@@ -151,6 +164,10 @@ menu .f0.f0.file.m.visit
 .f0.f0.file.m add command -label { Save gdb window } -command {save_gdb_window}
 .f0.f0.file.m add command -label { Save gdb init file } \
   -command {write_init_file [FileSelectBox -title "Select init file" -perm w]}
+.f0.f0.file.m add command -label { Add source directory } \
+  -command {add_source_dir}
+.f0.f0.file.m add command -label { Change working directory } \
+  -command {chg_directory}
 .f0.f0.file.m add separator
 .f0.f0.file.m add command -label { Quit } -command {exit_tgdb} \
   -accelerator {Ctrl-c}
@@ -171,6 +188,10 @@ set RaiseLabel " Auto-raise covered windows "
   -onvalue 1 -offvalue 0 -variable {Tgdb_option(UpdateBpts)}
 .f0.f0.options.m add checkbutton -label { Auto-save & load command history } \
   -onvalue 1 -offvalue 0 -variable {Tgdb_option(SaveHistory)}
+.f0.f0.options.m add checkbutton -label { cd to program path after loading } \
+  -onvalue 1 -offvalue 0 -variable {Tgdb_option(ProgPath)}
+.f0.f0.options.m add checkbutton -label { Remove CRs (^M) from source files } \
+  -onvalue 1 -offvalue 0 -variable {Tgdb_option(RemCRs)}
 .f0.f0.options.m add separator
 .f0.f0.options.m add checkbutton -label { Save preferences on exit } \
   -onvalue 1 -offvalue 0 -variable {Tgdb_option(SaveOptions)}
@@ -214,6 +235,8 @@ if { $debugger != "gdb166" } {
   -command {create_assembly_window}
 .f0.f0.windows.m add command -label { Xterm (shell) } -accelerator {Alt-x} \
   -command {create_shell_window}
+.f0.f0.windows.m add command -label { PXMon } -accelerator {Alt-o} \
+  -command {create_pxmon_window}
 .f0.f0.windows.m add separator
 .f0.f0.windows.m add command -label { Cycle windows } -accelerator {Alt-c} \
   -command {toggle_windows}
@@ -409,7 +432,7 @@ pack .f4.cont -side right
 pack .f4 -side top -fill x
 pack .f5.text -side left -fill both -expand 1
 pack .f5.scroll -side left -fill y
-pack .f5 -side top -fill both -expand 1
+pack .f5 -side top -fill both
 ################################################################################
 #
 # enable menu traversal
@@ -436,6 +459,7 @@ bind all <Alt-Any-Key-c> {toggle_windows}
 bind all <Alt-Any-Key-d> {create_memory_window}
 bind all <Alt-Any-Key-h> {create_tgdb_help_window}
 bind all <Alt-Any-Key-i> {help_about}
+bind all <Alt-Any-Key-o> {create_pxmon_window}
 bind all <Alt-Any-Key-r> {create_cpu_window}
 bind all <Alt-Any-Key-s> {create_stack_window}
 bind all <Alt-Any-Key-t> {raise_main_window}
@@ -452,6 +476,7 @@ bind .f3.text <Alt-Any-Key-c> {toggle_windows}
 bind .f3.text <Alt-Any-Key-d> {create_memory_window}
 bind .f3.text <Alt-Any-Key-h> {create_tgdb_help_window}
 bind .f3.text <Alt-Any-Key-i> {help_about}
+bind .f3.text <Alt-Any-Key-o> {create_pxmon_window}
 bind .f3.text <Alt-Any-Key-r> {create_cpu_window}
 bind .f3.text <Alt-Any-Key-s> {create_stack_window}
 bind .f3.text <Alt-Any-Key-t> {raise_main_window}
@@ -472,8 +497,6 @@ bind .f3.text <Motion> { set Line [int [%W index @%x,%y]] }
 bind .f3.text <B2-Motion> "[bind Text <B2-Motion>]"
 bind .f3.text <B2-Motion> "+[bind .f3.text <Motion>]"
 bind .f3.text <Double-2> {.f4.print invoke}
-#bind .f3.text <Any-Enter> { set old_focus(%W) [focus]; focus %W }
-#bind .f3.text <Any-Leave> { catch {focus $old_focus(%W)} }
 bind .f3.text <Up> {text_review [expr $Line - 2]}
 bind .f3.text <Down> {text_review $Line}
 bind .f3.text <Left> {
@@ -617,9 +640,19 @@ bind .f3.text <Shift-N> { search_string up }
 ################################################################################
 # these bindings handle text selection
 ################################################################################
-bind .f3.text <1> {select_c_expr %W %x %y}
-bind .f3.text <Double-1> {select_c_bracket_expr %W %x %y}
+bind .f3.text <1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
+  select_c_expr %W %x %y
+}
+bind .f3.text <Double-1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
+  select_c_bracket_expr %W %x %y
+}
 bind .f3.text <Triple-1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
   select_c_bracket_expr %W %x %y
   select_c_bracket_expr %W %x %y
 }
@@ -711,7 +744,7 @@ bind .f2.search <3> {+
 }
 ################################################################################
 #
-# bind status entry, i.e. delete all bindings except scanning and menu traversal
+# bind status entry
 #
 ################################################################################
 foreach bind [bind Entry] {
@@ -719,6 +752,9 @@ foreach bind [bind Entry] {
     bind .f1.status $bind { }
   }
 }
+bind .f1.status <1> {split_srcwin_init %W %x %y}
+bind .f1.status <B1-Motion> {split_srcwin_move %W %x %y}
+bind .f1.status <B1-ButtonRelease> {split_srcwin_end %W %x %y}
 ################################################################################
 #
 # bindings for help messages on buttons
@@ -795,6 +831,7 @@ bind .f5.text <Alt-Any-Key-c> {toggle_windows}
 bind .f5.text <Alt-Any-Key-d> {create_memory_window}
 bind .f5.text <Alt-Any-Key-h> {create_tgdb_help_window}
 bind .f5.text <Alt-Any-Key-i> {help_about}
+bind .f5.text <Alt-Any-Key-o> {create_pxmon_window}
 bind .f5.text <Alt-Any-Key-r> {create_cpu_window}
 bind .f5.text <Alt-Any-Key-s> {create_stack_window}
 bind .f5.text <Alt-Any-Key-t> {raise_main_window}
@@ -832,8 +869,8 @@ bind .f5.text <Return> {
       if { [info exists gdb_dont_repeat($xgdb_command)] } {
 	set gdb_command "echo"
       } elseif { [cequal $xgdb_command "x"] } {
-	# don't repeat args for last x command...
-	set gdb_command "x"
+	# don't repeat address args for last x command...
+	set gdb_command "x ?noarg?"
       }
     } else {
       set gdb_command "echo"
@@ -992,7 +1029,6 @@ bind .f5.text <Left> [bind .f5.text <Control-b>]
 bind .f5.text <Right> [bind .f5.text <Control-f>]
 bind .f5.text <Shift-Left> [bind .f5.text <Control-a>]
 bind .f5.text <Shift-Right> [bind .f5.text <Control-e>]
-bind .f5.text <1> {%W mark set anchor @%x,%y}
 bind .f5.text <Any-Enter> {
   if { [focus] == "none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
     set old_focus(f5.text) [focus]
@@ -1005,10 +1041,12 @@ bind .f5.text <Any-Leave> {
   }
 }
 bind .f5.text <Prior> {
-  %W yview [expr [int [%W index @0,0]] - [winfo height %W] / $WinSize(.,y)]
+  %W yview [expr [int [%W index @0,0]] - \
+    ([winfo height %W] - [lindex [%W conf -bd] 4]) / $WinSize(.,y2)]
 }
 bind .f5.text <Next> {
-  %W yview [expr [int [%W index @0,0]] + [winfo height %W] / $WinSize(.,y) - 2]
+  %W yview [expr [int [%W index @0,0]] -2 + \
+    ([winfo height %W] - [lindex [%W conf -bd] 4]) / $WinSize(.,y2)]
 }
 bind .f5.text <Shift-Up> [bind .f5.text <Prior>]
 bind .f5.text <Shift-Down> [bind .f5.text <Next>]
@@ -1018,12 +1056,19 @@ bind .f5.text <Tab> {command_line_completion}
 # these bindings handle text selection...
 ################################################################################
 bind .f5.text <1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
+  %W mark set anchor @%x,%y
   select_c_expr %W %x %y
 }
 bind .f5.text <Double-1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
   select_c_bracket_expr %W %x %y
 }
 bind .f5.text <Triple-1> {
+  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
+  bind %W <B1-Motion> { }
   select_c_bracket_expr %W %x %y
   select_c_bracket_expr %W %x %y
 }
@@ -1402,13 +1447,13 @@ proc blink_line {} {
 ################################################################################
 #
 # load a file and display it in the source text widget starting at line cur_line
-# and mark is line if highlight isn't "no"
+# and mark this line if highlight isn't "no"
 #
 ################################################################################
 proc load_file { name cur_line {highlight no} } {
   global Contents Line ThisFile TextBindings LentryBindings SearchBindings
   global Mark Marks Tags SearchLastLine LastFile LastLine colormodel
-  global BlinkingPC ModTime WinSize
+  global BlinkingPC ModTime WinSize Tgdb_option
 
   proc set_filemenu_state { state } {
     .f0.f0.file.m entryconfigure { Load program } -state $state
@@ -1420,6 +1465,8 @@ proc load_file { name cur_line {highlight no} } {
     .f0.f0.file.m entryconfigure { Visit } -state $state
     .f0.f0.file.m entryconfigure { Save gdb window } -state $state
     .f0.f0.file.m entryconfigure { Save gdb init file } -state $state
+    .f0.f0.file.m entryconfigure { Add source directory } -state $state
+    .f0.f0.file.m entryconfigure { Change working directory } -state $state
   }
 
   if { [cequal $name ""] } {
@@ -1428,6 +1475,9 @@ proc load_file { name cur_line {highlight no} } {
     }
     return
   }
+  if { ![file readable $name] } {
+    set name [file2nfspath $name]
+  }
   if { [cequal $name $ThisFile] && ([file mtime $name] > $ModTime($name)) } {
     # reload current file
     show_status "Reloading file \"$name\"..." steady
@@ -1447,6 +1497,10 @@ proc load_file { name cur_line {highlight no} } {
       return
     }
     close $desc
+    if { $Tgdb_option(RemCRs) } {
+      regsub -all "
+\n" $Contents($name) "\n" Contents($name)
+    }
     set ModTime($name) [file mtime $name]
     .f3.text configure -state normal
     .f3.text delete 1.0 end
@@ -1466,7 +1520,9 @@ proc load_file { name cur_line {highlight no} } {
   if { [cequal $name $ThisFile] } {
     # the following lines provide a better "pickplace" behaviour...
     set topline [int [.f3.text index @0,0]]
-    set lstline [expr $topline + ([winfo height .f3.text] / $WinSize(.,y)) - 1]
+    set lstline [expr $topline - 1 \
+      + (([winfo height .f3.text] - [lindex [.f3.text conf -bd] 4]) \
+      / $WinSize(.,y))]
     if { ($cur_line <= $topline) && ($cur_line >= [expr $topline - 5]) } {
       set Line [expr $topline -8]; adjust_curline exact
     } elseif { ($cur_line >= $lstline) && ($cur_line <= [expr $lstline + 5]) } {
@@ -1523,6 +1579,10 @@ proc load_file { name cur_line {highlight no} } {
       return
     }
     close $desc
+    if { $Tgdb_option(RemCRs) } {
+      regsub -all "
+\n" $Contents($name) "\n" Contents($name)
+    }
     set ModTime($name) [file mtime $name]
   }
 
@@ -1896,6 +1956,38 @@ proc save_gdb_window {} {
   }
 }
 ################################################################################
+# add a directory to the search path for source files
+################################################################################
+proc add_source_dir {} {
+  global Tgdb_cmd
+
+  set file [FileSelectBox -title "Add directory to source path" -file .]
+  if { ![cequal $file ""] } {
+    if { ![file isdirectory $file] } {
+      show_status "No such directory: \"$file\"" steady
+      bell
+    } else {
+      $Tgdb_cmd(directory) $file
+    }
+  }
+}
+################################################################################
+# change to directory
+################################################################################
+proc chg_directory {} {
+  global Tgdb_cmd
+
+  set file [FileSelectBox -title "Add directory to source path" -file .]
+  if { ![cequal $file ""] } {
+    if { ![file isdirectory $file] } {
+      show_status "No such directory: \"$file\"" steady
+      bell
+    } else {
+      $Tgdb_cmd(cd) $file
+    }
+  }
+}
+################################################################################
 # see if there is a breakpoint at the line denoted by x,y in window w; if so,
 # tell gdb to delete it, otherwise tell him to set one at the first line that
 # is known to contain any code
@@ -1998,7 +2090,7 @@ proc busy_create {} {
   foreach sibling [winfo children .] {
     if { $sibling == ".f4" } {
       foreach but [winfo children .f4] {
-	blt_busy hold $but
+ 	blt_busy hold $but
       }
       blt_busy forget .f4.stop
     } else {
@@ -2108,7 +2200,7 @@ proc do_dialog {say_what how {showpc no} {loadfile yes} } {
 	continue
       }
       if { $Tgdb_interactive } {
-	if { ![regexp {[^:]*:[0-9]+:([0-9]+):} $line] } {
+	if { ![regexp {(.:)?[^:]*[:;][0-9]+[:;]([0-9]+)[:;]} $line]} {
 	  .f5.text insert end "$line\n"
 	  .f5.text yview -pickplace end
 	  update idletasks
@@ -2119,7 +2211,7 @@ proc do_dialog {say_what how {showpc no} {loadfile yes} } {
       if {   [regexp "^The program is not" $line]
           || [regexp "^No selected frame" $line]
           || [regexp "^Value returned " $line]
-	  || [regexp {[^:]*:[0-9]+:([0-9]+):} $line]
+	  || [regexp {(.:)?[^:]*[:;][0-9]+[:;]([0-9]+)[:;]} $line]
 	  || [regexp {^"finish" not meaningful} $line] } {
         append result "$line\n"
 	continue
@@ -2235,6 +2327,7 @@ proc do_dialog {say_what how {showpc no} {loadfile yes} } {
 	append QueryResult n
 	exp_send "n\r"
       }
+      set first 1
       set filter_query 1
       set answer ""
       exp_continue
@@ -2334,11 +2427,14 @@ proc do_dialog {say_what how {showpc no} {loadfile yes} } {
     # don't display any sources if we stopped at a
     # place where no debug info is available...
     if {   ($showpc != "no" )
-	&& ![regexp {[^:]*:[0-9]+:[0-9]+:(beg|middle):0x[0-9a-f]+} $answer]
+	&& ![regexp \
+	      {(.:)?[^:]*[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-f]+} \
+	      $answer]
        } {
       exp_send "info line *\$pc\r"
       expect $prompt {
-	if {   ![regexp {[^:]*:[0-9]+:[0-9]+:(beg|middle):0x[0-9a-f]+} \
+	if {   ![regexp \
+	       {(.:)?[^:]*[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-f]+}\
 	          $expect_out(buffer)]
 	    && ![regexp {Line [0-9]+ of "(.*)" .* at (pc|address) 0x[0-9a-f]+} \
 		  $expect_out(buffer)] } {
@@ -2356,15 +2452,17 @@ proc do_dialog {say_what how {showpc no} {loadfile yes} } {
         }
       }
     }
-    while { [regexp {([^:]*):([0-9]+):([0-9]+):(beg|middle):(0x[0-9a-f]+)} \
-	      $answer {} file line off pc] } {
+    while { [regexp \
+       {((.:)?[^:]*)[:;]([0-9]+)[:;]([0-9]+)[:;](beg|middle):(0x[0-9a-f]+)}\
+	      $answer {} file {} line off pc] } {
       if { $showpc == "no" } {
         load_file $file $line no
       } else {
         load_file $file $line yes
       }
-      regsub {[^:]*:[0-9]+:[0-9]+:(beg|middle):0x[0-9a-fA-F]*.} $answer \
-        "" answer
+      regsub \
+	{(.:)?[^:]*[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-fA-F]*.} \
+	$answer "" answer
     }
   }
   regsub -all "Reading in symbols for .*done.\n" $answer "" answer
@@ -2410,7 +2508,7 @@ proc filter_output { cmd {arg ""} } {
 ################################################################################
 proc exit_tgdb {} {
   global xterm_pid QueryResult ExitTgdb
-  global Tgdb_option WinPos Tgdb_startfile WinProc AsmOpts
+  global Tgdb_option WinPos WinSize Tgdb_startfile WinProc AsmOpts
   global gdb_history gdb_history_nr gdb_history_max gdb_history_spot
 
   # avoid nested calling (due to EOF condition when killing gdb in do_dialog)
@@ -2458,11 +2556,16 @@ proc exit_tgdb {} {
       foreach win [array names WinPos] {
 	puts $desc "set WinPos($win) \"$WinPos($win)\""
       }
+      # remember gdb window's height
+      set txtb [lindex [.f5.text conf -bd] 4]
+      set txth $WinSize(.,y2)
+      puts $desc "set WinGdb [expr ([winfo height .f5.text] - $txtb) / $txth]"
       # remember which windows are currently opened
       if { $windows != "" } {
         puts $desc "\n# open windows"
 	foreach win $windows {
-	  if { [lsearch -exact ". .help .dbg .thelp" $win] != -1 } continue
+	  if { [lsearch -exact ". .help .dbg .thelp .pxmon" $win] != -1 } \
+	    continue
 	  puts $desc "catch \{$WinProc($win)\}"
 	}
       }
@@ -2492,25 +2595,34 @@ proc exit_tgdb {} {
 ################################################################################
 proc init_srcwin {} {
   global Mark Line ThisFile Contents TextBindings LentryBindings SearchBindings
-  global LastFile LastFile colormodel
+  global LastFile LastLine colormodel
 
+  if { [cequal [bind .f3.text <1>] " "] } {
+    set save_bindings 0
+  } else {
+    set save_bindings 1
+  }
+  .f3.text configure -state normal
   .f3.text delete 1.0 end
+  .f3.text configure -state disabled
   catch {unset Contents}
   set ThisFile ""
   set Line ""
   set LastFile ""
   set LastLine 0
-  foreach bind [bind .f3.text] {
-    set TextBindings($bind) [bind .f3.text $bind]
-    bind .f3.text $bind { }
-  }
-  foreach bind [bind .title.f1.lentry] {
-    set LentryBindings($bind) [bind .title.f1.lentry $bind]
-    bind .title.f1.lentry $bind { }
-  }
-  foreach bind [bind .f2.search] {
-    set SearchBindings($bind) [bind .f2.search $bind]
-    bind .f2.search $bind { }
+  if { $save_bindings } {
+    foreach bind [bind .f3.text] {
+      set TextBindings($bind) [bind .f3.text $bind]
+      bind .f3.text $bind { }
+    }
+    foreach bind [bind .title.f1.lentry] {
+      set LentryBindings($bind) [bind .title.f1.lentry $bind]
+      bind .title.f1.lentry $bind { }
+    }
+    foreach bind [bind .f2.search] {
+      set SearchBindings($bind) [bind .f2.search $bind]
+      bind .f2.search $bind { }
+    }
   }
   if { $colormodel == "color" } {
     .f3.text tag configure mytag -background green
@@ -2536,6 +2648,8 @@ set gdb_history_max 128
 set gdb_history(0) ""
 set gdb_history_nr 0
 set gdb_history_spot 0
+set Tgdb_option(RemCRs) 0
+set Tgdb_option(ProgPath) 0
 set Tgdb_option(SaveHistory) 1
 set Tgdb_option(bell) visual
 set Tgdb_option(bell,active) 0
@@ -2558,12 +2672,19 @@ set timeout -1
 #
 ################################################################################
 if { "$debugger" == "gdb166" } {
-  set gdb_invocation "spawn -noecho $debugger -nx -fullname [split $gdb_argv]"
-  eval $gdb_invocation
+  set gdb_invocation "$debugger -nx -fullname [split $gdb_argv]"
+  eval spawn -noecho $dbgrsh $gdb_invocation
   do_dialog "" silent
-  do_dialog "set screen 0" silent
-  do_dialog "set array-max 0" silent
-  do_dialog "set pretty" silent
+  if { $gdb16x } {
+    do_dialog "set height 0" silent
+    do_dialog "set width 0" silent
+    do_dialog "set print elements 0" silent
+    do_dialog "set print pretty" silent
+  } else {
+    do_dialog "set screen 0 0" silent
+    do_dialog "set array-max 0" silent
+    do_dialog "set pretty" silent
+  }
 } else {
   toplevel .dbg -cursor top_left_arrow
   wm withdraw .dbg
@@ -2584,7 +2705,11 @@ if { "$debugger" == "gdb166" } {
   set xterm_line [gets $xterm_desc]
   set xterm_xid [lindex $xterm_line 0]
   set xterm_tty [lindex $xterm_line 1]
-  set xterm_pid [lindex $xterm_line 2]
+  if { ![scan [lindex $xterm_line 2] "%d" xterm_pid] } {
+    puts $std_out "Couldn't read xterm_pid (maybe a problem with mkxtty?)."
+    catch {.shell.f0.xterm stoptksteal}
+    exit
+  } 
   catch {close $xterm_desc}
   tksteal .dbg.f0.xterm -name "Tgdb debugee" -windowid $xterm_xid
   pack .dbg.f0 -expand 1 -fill both
@@ -2596,12 +2721,13 @@ if { "$debugger" == "gdb166" } {
   catch {set orig_shell $env(SHELL)}
   set env(SHELL) callcld
   set env(TGDB_TTYDEV) $xterm_tty
-  set gdb_invocation "spawn -noecho $debugger -nx -fullname [split $gdb_argv]"
-  eval $gdb_invocation
+  set gdb_invocation "$debugger -nx -fullname [split $gdb_argv]"
+  eval spawn -noecho $dbgrsh $gdb_invocation
   catch {set env(SHELL) $orig_shell}
   unset env(TGDB_TTYDEV)
   do_dialog "" silent
   do_dialog "set height 0" silent
+  do_dialog "set width 0" silent
   do_dialog "set print elements 0" silent
   do_dialog "set print pretty" silent
   unset xterm_desc xterm_line debcolor debfont orig_shell
@@ -2612,16 +2738,21 @@ destroy .welcome
 ################################################################################
 #
 # try to load tgdb's startup file "./.t"$debugger"init" or "~/.t"$debugger"init"
+# (or as specified in the TGDB_INITFILE environment variable)
 #
 ################################################################################
-set startfile0 ""; append startfile0 "./.t" $debugger "init"
-set startfile1 ""; append startfile1 "~/.t" $debugger "init"
-if { [file readable $startfile0] } {
-  set Tgdb_startfile $startfile0
+if { [info exists env(TGDB_INITFILE)] } {
+  set Tgdb_startfile $env(TGDB_INITFILE)
 } else {
-  set Tgdb_startfile $startfile1
+  set startfile0 ""; append startfile0 "./.t" $debugger "init"
+  set startfile1 ""; append startfile1 "~/.t" $debugger "init"
+  if { [file readable $startfile0] } {
+    set Tgdb_startfile $startfile0
+  } else {
+    set Tgdb_startfile $startfile1
+  }
+  unset startfile0 startfile1
 }
-unset startfile0 startfile1
 catch {source $Tgdb_startfile}
 
 ################################################################################
@@ -2657,18 +2788,26 @@ wm protocol . WM_TAKE_FOCUS {focus .f5.text}
 wm protocol . WM_DELETE_WINDOW {exit_tgdb}
 bind . <Visibility> {check_visibility %W}
 if { [catch {wm geometry . $WinPos(.)}] } {
-  wm geometry . +100+100
+  wm geometry . 80x15+100+100
 }
-wm minsize . 40 15
+wm minsize . 80 5
 wm deiconify .
 focus none
 tkwait visibility .
 update
-set txtw [lindex [.f3.text configure -width] 4]
-set txth [lindex [.f3.text configure -height] 4]
-set WinSize(.,x) [expr [winfo width .f3.text] / $txtw]
-set WinSize(.,y) [expr [winfo height .f3.text] / $txth]
-unset txtw txth
+set txtb [lindex [.f3.text conf -bd] 4]
+scan [wm geometry .] "%dx%d" txtw txth
+set WinSize(.,x) [expr ([winfo width .f3.text] - $txtb) / $txtw]
+set WinSize(.,y) [expr ([winfo height .f3.text] - $txtb) / $txth]
+set txtw [lindex [.f5.text conf -width] 4]
+set txtb [lindex [.f5.text conf -bd] 4]
+set txth [lindex [.f5.text conf -height] 4]
+set WinSize(.,x2) [expr ([winfo width .f5.text] - $txtb) / $txtw]
+set WinSize(.,y2) [expr ([winfo height .f5.text] - $txtb) / $txth]
+unset txtw txtb txth
+if { [info exists WinGdb] } {
+  .f5.text configure -height $WinGdb
+}
 
 ################################################################################
 #
@@ -2682,7 +2821,7 @@ init_gdb_help
 init_tgdb_cmds
 show_copyright_and_license_policy
 
-if { "$debugger" == "gdb166" } {
+if { ("$debugger" == "gdb166") && !$gdb16x } {
   filter_output "info version"
 } else {
   filter_output "show version"
@@ -2693,6 +2832,13 @@ if { "$debugger" == "gdb166" } {
 # and list main() function if no core file was specified
 #
 ################################################################################
+if { $Tgdb_option(ProgPath) } {
+  if { ![cequal $gdb_symfile ""] } {
+    set Tgdb_cwd [file dirname $gdb_symfile]
+    $Tgdb_cmd(cd) $Tgdb_cwd
+  }
+}
+$Tgdb_cmd(directory) $Tgdb_cwd
 foreach file "$gdb_execfile $gdb_symfile $gdb_corefile" {
   set path [file dirname $file]
   if { ![cequal $path ""] && ![cequal $path "."] } {
